# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for calculating the mass matrix."""
import itertools

from brax.v2 import scan
from brax.v2.base import System
import jax
from jax import numpy as jp


def matrix(sys: System, cinr: jp.ndarray, cdof: jp.ndarray) -> jp.ndarray:
  """Calculates the mass matrix for the system given joint position.

  This function uses the Composite-Rigid-Body Algorithm as described here:

  https://users.dimi.uniud.it/~antonio.dangelo/Robotica/2019/helper/Handbook-dynamics.pdf

  Args:
    sys: system defining the kinematic tree and other properties
    cinr: inertia in com frame
    cdof: dofs in com frame

  Returns:
    a symmetric positive matrix (qd_size, qd_size) representing the generalized
    mass and inertia of the system
  """
  # backward scan up tree: accumulate composite link inertias
  def crb_fn(crb_child, crb):
    if crb_child is not None:
      crb += crb_child
    return crb

  crb = scan.tree(sys, crb_fn, 'l', cinr, reverse=True)

  # expand composite inertias to a matrix: M[i,j] = cdof_j * crb[i] * cdof_i
  @jax.vmap
  def mx_row(dof_link, cdof_i):
    f = crb.take(dof_link).mul(cdof_i)

    @jax.vmap
    def mx_col(cdof_j):
      return cdof_j.dot(f)

    return mx_col(cdof)

  mx = mx_row(sys.dof_link(), cdof)

  # mask out empty parts of the matrix
  si, sj = [], []
  dof_ranges = sys.dof_ranges()
  for i in range(len(sys.link_parents)):
    j = i
    while j > -1:
      for dof_i, dof_j in itertools.product(dof_ranges[i], dof_ranges[j]):
        si, sj = si + [dof_i], sj + [dof_j]
      j = sys.link_parents[j]

  mx = mx * jp.zeros_like(mx).at[(jp.array(si), jp.array(sj))].set(1.0)

  # we mask i, j<=i, which is the lower triangular portion of the matrix
  # mirror it onto the upper triangular
  mx = jp.tril(mx) + jp.tril(mx, -1).T

  # add the armature inertia for rotors
  mx = mx + jp.diag(sys.dof.armature)

  return mx



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functionality for brax bodies."""

from brax import jumpy as jp
from brax import math
from brax import pytree
from brax.physics import config_pb2
from brax.physics.base import P, QP, vec_to_arr


@pytree.register
class Body:
  """A body is a solid, non-deformable object with some mass and shape.

  Attributes:
    idx: Index of where body is found in the system.
    inertia: (3, 3) Inverse Inertia matrix represented in body frame.
    mass: Mass of the body.
    active: whether the body is effected by physics calculations
    index: name->index dict for looking up body names
  """
  __pytree_ignore__ = ('index',)

  def __init__(self, config: config_pb2.Config):
    self.idx = jp.arange(0, len(config.bodies))
    self.inertia = 1. / jp.array([vec_to_arr(b.inertia) for b in config.bodies])
    self.mass = jp.array([b.mass for b in config.bodies])
    self.active = jp.array(
        [0.0 if b.frozen.all else 1.0 for b in config.bodies])
    self.index = {b.name: i for i, b in enumerate(config.bodies)}

  def impulse(self, qp: QP, impulse: jp.ndarray, pos: jp.ndarray) -> P:
    """Calculates updates to state information based on an impulse.

    Args:
      qp: State data of the system
      impulse: Impulse vector
      pos: Location of the impulse relative to the body's center of mass

    Returns:
      dP: An impulse to apply to this body
    """
    dvel = impulse / self.mass
    dang = self.inertia * jp.cross(pos - qp.pos, impulse)
    return P(vel=dvel, ang=dang)


def min_z(qp: QP, body: config_pb2.Body) -> float:
  """Returns the lowest z of all the colliders in a body."""
  if not body.colliders:
    return 0.0

  result = float('inf')

  for col in body.colliders:
    if col.HasField('sphere'):
      sphere_pos = math.rotate(vec_to_arr(col.position), qp.rot)
      z = qp.pos[2] + sphere_pos[2] - col.sphere.radius
      result = jp.amin(jp.array([result, z]))
    elif col.HasField('capsule'):
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      axis = math.rotate(jp.array([0., 0., 1.]), rot)
      length = col.capsule.length / 2 - col.capsule.radius
      for end in (-1, 1):
        sphere_pos = vec_to_arr(col.position) + end * axis * length
        sphere_pos = math.rotate(sphere_pos, qp.rot)
        z = qp.pos[2] + sphere_pos[2] - col.capsule.radius
        result = jp.amin(jp.array([result, z]))
    elif col.HasField('box'):
      corners = [(i % 2 * 2 - 1, 2 * (i // 4) - 1, i // 2 % 2 * 2 - 1)
                 for i in range(8)]
      corners = jp.array(corners, dtype=float)
      for corner in corners:
        corner = corner * vec_to_arr(col.box.halfsize)
        rot = math.euler_to_quat(vec_to_arr(col.rotation))
        corner = math.rotate(corner, rot)
        corner = corner + vec_to_arr(col.position)
        corner = math.rotate(corner, qp.rot) + qp.pos
        result = jp.amin(jp.array([result, corner[2]]))
    else:
      # ignore planes and other stuff
      result = jp.amin(jp.array([result, 0.0]))

  return result



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for integrating maximal coordinate dynamics."""
# pylint:disable=g-multiple-import
from typing import Optional, Tuple

from brax.v2 import math
from brax.v2.base import Motion, System, Transform
from brax.v2.spring import maximal
import jax
from jax import numpy as jp


def _kinetic(sys: System, x: Transform, xd: Motion) -> Transform:
  """Performs a kinetic integration step.

  Args:
    sys: System to integrate
    x: Current world Transform
    xd: Current world Motion

  Returns:
    Position state integrated with current velocity state.
  """

  @jax.vmap
  def op(x: Transform, xd: Motion) -> Transform:
    pos = x.pos + xd.vel * sys.dt
    rot_at_ang_quat = math.ang_to_quat(xd.ang) * 0.5 * sys.dt
    rot = x.rot + math.quat_mul(rot_at_ang_quat, x.rot)
    rot = rot / jp.linalg.norm(rot)
    return Transform(pos=pos, rot=rot)

  return op(x, xd)


def _potential(
    sys: System,
    x: Transform,
    xd: Motion,
    dxdd: Optional[Motion] = None,
    dxd: Optional[Motion] = None,
    dx: Optional[Transform] = None,
) -> Motion:
  """Performs an arg dependent integrator step.

  Args:
    sys: System to integrate
    x: Current world Transform
    xd: Current world Motion
    dxdd: Acceleration level update
    dxd: Velocity level update
    dx: Position level update

  Returns:
    State data advanced by one potential integration step.
  """

  del x, dx

  @jax.vmap
  def op_acc(xd, dxdd) -> Motion:
    vel = jp.exp(sys.vel_damping * sys.dt) * xd.vel
    vel += (dxdd.vel + sys.gravity) * sys.dt
    ang = jp.exp(sys.ang_damping * sys.dt) * xd.ang
    ang += dxdd.ang * sys.dt
    return Motion(vel=vel, ang=ang)

  # TODO: fill-in position level updates

  if dxd is not None:
    return xd + dxd
  elif dxdd is not None:
    return op_acc(xd, dxdd)
  return xd


def forward(
    sys: System,
    xi: Transform,
    xdi: Motion,
    inv_inertia: jp.ndarray,
    f: Motion,
    pos: jp.ndarray,
    link_idx: jp.ndarray,
) -> Motion:
  """Updates state with forces in the center of mass frame.

  Args:
    sys: System to forward propagate
    xi: Transform state of links to update in maximal CoM coordinates
    xdi: Motion state of links to update in maximal CoM coordinates
    inv_inertia: Inverse inertia matrix at the CoM in world frame
    f: Motion with dim (m, 3) with forces to apply to links
    pos: (m, 3) ndarray of world-frame locations to apply force
    link_idx: (m,) ndarray of link indices which receive `f` forces

  Returns:
    Updated state Motion.
  """
  # apply forces
  xi_all = xi.take(link_idx)
  links = sys.link.take(link_idx)
  inv_inertia = inv_inertia.take(link_idx, axis=0)
  xddi_all = jax.vmap(maximal.world_impulse)(
      xi_all.pos, links.inertia.mass, inv_inertia, f.vel, pos, torque=f.ang
  )
  # sum forces over links
  dxddi = jax.tree_map(
      lambda f: jax.ops.segment_sum(f, link_idx, xi.pos.shape[0]), xddi_all
  )
  xdi = _potential(sys, xi, xdi, dxdd=dxddi)
  return xdi


def forward_c(
    sys: System,
    xi: Transform,
    xdi: Motion,
    inv_inertia: jp.ndarray,
    p: Motion,
    pos: jp.ndarray,
    link_idx: jp.ndarray,
) -> Tuple[Transform, Motion]:
  """Updates state with velocity update in the center of mass frame.

  Args:
    sys: System to forward propagate
    xi: Transform state of links to update in maximal CoM coordinates
    xdi: Motion state of links to update in maximal CoM coordinates
    inv_inertia: Inverse inertia matrix at the CoM in world frame
    p: Motion with dim (m, 3) of impulses to apply to links
    pos: (m, 3) ndarray of world-frame locations to apply impulse
    link_idx: (m,) ndarray of link indices which receive impulse

  Returns:
    Updated state Transform and Motion.
  """
  # transform into center of mass frame
  if link_idx.shape[0] != 0:
    # apply velocity update `p`
    xi_all = xi.take(link_idx)
    links = sys.link.take(link_idx)
    inv_inertia = inv_inertia.take(link_idx, axis=0)
    xdi_all = jax.vmap(maximal.world_impulse)(
        xi_all.pos, links.inertia.mass, inv_inertia, p.vel, pos, torque=None
    )
    # average xd over links
    apply_v = jp.where(jp.any(p.vel, axis=-1) & (link_idx != -1), 1.0, 0.0)
    n_v = jax.ops.segment_sum(apply_v, link_idx, sys.num_links())
    n_v = jp.reshape(1e-8 + n_v, (sys.num_links(), 1))
    dxdi = jax.tree_map(
        lambda p: jax.ops.segment_sum(  # pylint:disable=g-long-lambda
            p, link_idx, xi.pos.shape[0]
        )
        / n_v,
        xdi_all,
    )
    xdi = _potential(sys, xi, xdi, dxd=dxdi)

  xi = _kinetic(sys, xi, xdi)
  return xi, xdi



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for physics calculations in maximal coordinates."""
# pylint:disable=g-multiple-import
from typing import Tuple, Optional

from brax.v2 import math
from brax.v2.base import Motion, System, Transform
import jax
from jax import numpy as jp


def maximal_to_com(
    sys: System, x: Transform, xd: Motion
) -> Tuple[Transform, Motion]:
  """Translates link-frame Transforms and Motions into CoM-frame."""
  xi = x.vmap().do(sys.link.inertia.transform)
  com_transform = Transform(pos=xi.pos, rot=x.rot)
  com_motion = Motion(
      vel=xd.vel + jax.vmap(jp.cross)(xd.ang, xi.pos - x.pos), ang=xd.ang
  )
  return com_transform, com_motion


def com_to_maximal(
    x_com: Transform, xd_com: Motion, coord_transform: Transform
) -> Tuple[Transform, Motion]:
  """Translates CoM-frame Transforms and Motions into link-frame."""

  final_shift = jax.vmap(math.rotate)(
      jax.vmap(math.rotate)(
          coord_transform.pos, math.quat_inv(coord_transform.rot)
      ),
      x_com.rot,
  )

  maximal_transform = Transform(pos=x_com.pos - final_shift, rot=x_com.rot)
  maximal_motion = Motion(
      vel=xd_com.vel - jax.vmap(jp.cross)(xd_com.ang, final_shift),
      ang=xd_com.ang,
  )

  return maximal_transform, maximal_motion


def com_inv_inertia(sys, x) -> jp.ndarray:
  """Gets the inverse inertia at the center of mass in world frame."""

  @jax.vmap
  def inv_i(link_inertia, x_rot):
    r_inv = math.quat_inv(link_inertia.transform.rot)
    ri = math.quat_mul(r_inv, x_rot)
    i_rot_row = jax.vmap(math.rotate, in_axes=[0, None])(link_inertia.i, ri)
    i_rot_col = jax.vmap(math.rotate, in_axes=[0, None])(i_rot_row.T, ri)
    return math.inv_3x3(i_rot_col)

  return inv_i(sys.link.inertia, x.rot)


def to_world(x, xd, rpos: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
  """Returns world information about a point relative to a part.

  Args:
    x: Transform world
    xd: Motion world
    rpos: Point relative to center of mass of part.

  Returns:
      world_pos: World-space coordinates of rpos
      world_vel: World-space velocity of rpos
  """
  rpos_off = math.rotate(rpos, x.rot)
  rvel = jp.cross(xd.ang, rpos_off)
  return x.pos + rpos_off, xd.vel + rvel


def world_impulse(
    x,
    mass,
    inv_inertia,
    impulse: jp.ndarray,
    pos: jp.ndarray,
    torque: Optional[jp.ndarray],
) -> Motion:
  """Calculates updates to world state information based on an impulse.

  Args:
    x: world position
    mass: body mass
    inv_inertia: body inverse inertia
    impulse: impulse vector
    pos: location of the impulse relative to the body's center of mass
    torque: additional torque to apply

  Returns:
    Change in velocity in world space
  """
  if torque is None:
    torque = jp.zeros(3)
  dvel = impulse / mass
  dang = inv_inertia @ (jp.cross(pos - x, impulse) + torque)
  return Motion(vel=dvel, ang=dang)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Analytic policy gradient training."""

import functools
import time
from typing import Callable, Optional, Tuple

from absl import logging
from brax import envs
from brax.envs import wrappers
from brax.training import acting
from brax.training import pmap
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.apg import networks as apg_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
import flax
import jax
import jax.numpy as jnp
import optax

InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]
Metrics = types.Metrics

_PMAP_AXIS_NAME = 'i'


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  normalizer_params: running_statistics.RunningStatisticsState
  policy_params: Params


def _unpmap(v):
  return jax.tree_util.tree_map(lambda x: x[0], v)


def train(environment: envs.Env,
          episode_length: int,
          action_repeat: int = 1,
          num_envs: int = 1,
          max_devices_per_host: Optional[int] = None,
          num_eval_envs: int = 128,
          learning_rate: float = 1e-4,
          seed: int = 0,
          truncation_length: Optional[int] = None,
          max_gradient_norm: float = 1e9,
          num_evals: int = 1,
          normalize_observations: bool = False,
          deterministic_eval: bool = False,
          network_factory: types.NetworkFactory[
              apg_networks.APGNetworks] = apg_networks.make_apg_networks,
          progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
          eval_env: Optional[envs.Env] = None):
  """Direct trajectory optimization training."""
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)
  device_count = local_devices_to_use * process_count

  if truncation_length is not None:
    assert truncation_length > 0

  num_evals_after_init = max(num_evals - 1, 1)

  assert num_envs % device_count == 0
  env = environment
  env = wrappers.wrap_for_training(
      env, episode_length=episode_length, action_repeat=action_repeat)

  normalize = lambda x, y: x
  if normalize_observations:
    normalize = running_statistics.normalize
  apg_network = network_factory(
      env.observation_size,
      env.action_size,
      preprocess_observations_fn=normalize)
  make_policy = apg_networks.make_inference_fn(apg_network)

  optimizer = optax.adam(learning_rate=learning_rate)

  def env_step(carry: Tuple[envs.State, PRNGKey], step_index: int,
               policy: types.Policy):
    env_state, key = carry
    key, key_sample = jax.random.split(key)
    actions = policy(env_state.obs, key_sample)[0]
    nstate = env.step(env_state, actions)
    if truncation_length is not None:
      nstate = jax.lax.cond(
          jnp.mod(step_index + 1, truncation_length) == 0.,
          jax.lax.stop_gradient, lambda x: x, nstate)

    return (nstate, key), (nstate.reward, env_state.obs)

  def loss(policy_params, normalizer_params, key):
    key_reset, key_scan = jax.random.split(key)
    env_state = env.reset(
        jax.random.split(key_reset, num_envs // process_count))
    f = functools.partial(
        env_step, policy=make_policy((normalizer_params, policy_params)))
    (rewards,
     obs) = jax.lax.scan(f, (env_state, key_scan),
                         (jnp.array(range(episode_length // action_repeat))))[1]
    return -jnp.mean(rewards), obs

  loss_grad = jax.grad(loss, has_aux=True)

  def clip_by_global_norm(updates):
    g_norm = optax.global_norm(updates)
    trigger = g_norm < max_gradient_norm
    return jax.tree_util.tree_map(
        lambda t: jnp.where(trigger, t, (t / g_norm) * max_gradient_norm),
        updates)

  def training_epoch(training_state: TrainingState, key: PRNGKey):
    key, key_grad = jax.random.split(key)
    grad, obs = loss_grad(training_state.policy_params,
                          training_state.normalizer_params, key_grad)
    grad = clip_by_global_norm(grad)
    grad = jax.lax.pmean(grad, axis_name='i')
    params_update, optimizer_state = optimizer.update(
        grad, training_state.optimizer_state)
    policy_params = optax.apply_updates(training_state.policy_params,
                                        params_update)

    normalizer_params = running_statistics.update(
        training_state.normalizer_params, obs, pmap_axis_name=_PMAP_AXIS_NAME)

    metrics = {
        'grad_norm': optax.global_norm(grad),
        'params_norm': optax.global_norm(policy_params)
    }
    return TrainingState(
        optimizer_state=optimizer_state,
        normalizer_params=normalizer_params,
        policy_params=policy_params), metrics

  training_epoch = jax.pmap(training_epoch, axis_name=_PMAP_AXIS_NAME)

  training_walltime = 0

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(training_state: TrainingState,
                                 key: PRNGKey) -> Tuple[TrainingState, Metrics]:
    nonlocal training_walltime
    t = time.time()
    (training_state, metrics) = training_epoch(training_state, key)
    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (episode_length * num_envs) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()}
    }
    return training_state, metrics

  key = jax.random.PRNGKey(seed)
  global_key, local_key = jax.random.split(key)
  del key
  local_key = jax.random.fold_in(local_key, process_id)
  local_key, eval_key = jax.random.split(local_key)

  # The network key should be global, so that networks are initialized the same
  # way for different processes.
  policy_params = apg_network.policy_network.init(global_key)
  del global_key

  training_state = TrainingState(
      optimizer_state=optimizer.init(policy_params),
      policy_params=policy_params,
      normalizer_params=running_statistics.init_state(
          specs.Array((env.observation_size,), jnp.float32)))
  training_state = jax.device_put_replicated(
      training_state,
      jax.local_devices()[:local_devices_to_use])

  if not eval_env:
    eval_env = env
  else:
    eval_env = wrappers.wrap_for_training(
        eval_env, episode_length=episode_length, action_repeat=action_repeat)

  evaluator = acting.Evaluator(
      eval_env,
      functools.partial(make_policy, deterministic=deterministic_eval),
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key)

  # Run initial eval
  if process_id == 0 and num_evals > 1:
    metrics = evaluator.run_evaluation(
        _unpmap(
            (training_state.normalizer_params, training_state.policy_params)),
        training_metrics={})
    logging.info(metrics)
    progress_fn(0, metrics)

  for it in range(num_evals_after_init):
    logging.info('starting iteration %s %s', it, time.time() - xt)

    # optimization
    epoch_key, local_key = jax.random.split(local_key)
    epoch_keys = jax.random.split(epoch_key, local_devices_to_use)
    (training_state,
     training_metrics) = training_epoch_with_timing(training_state, epoch_keys)

    if process_id == 0:
      # Run evals.
      metrics = evaluator.run_evaluation(
          _unpmap(
              (training_state.normalizer_params, training_state.policy_params)),
          training_metrics)
      logging.info(metrics)
      progress_fn(it + 1, metrics)

  # If there was no mistakes the training_state should still be identical on all
  # devices.
  pmap.assert_is_replicated(training_state)
  params = _unpmap(
      (training_state.normalizer_params, training_state.policy_params))
  pmap.synchronize_hosts()
  return (make_policy, params, metrics)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for constraints."""

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import test_utils
from brax.v2.generalized import pipeline
import jax
import numpy as np


class ConstraintTest(parameterized.TestCase):

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('halfcheetah.xml',),
  )
  def test_jacobian(self, xml_file):
    """Test constraint jacobian."""
    sys = test_utils.load_fixture(xml_file)
    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)

      # ignore zero rows (brax jacobians are static size)
      np.testing.assert_almost_equal(
          np.sort(state.con_jac[np.any(state.con_jac, axis=1)], axis=0),
          np.sort(mj_next.efc_J, axis=0),
          6,
      )
      np.testing.assert_almost_equal(
          np.sort(state.con_pos[np.any(state.con_jac, axis=1)], axis=0),
          np.sort(mj_next.efc_pos, axis=0),
          6,
      )

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('halfcheetah.xml',),
  )
  def test_force(self, xml_file):
    """Test constraint force."""
    sys = test_utils.load_fixture(xml_file)
    # brax can offer constraint solutions that are just as good as mujoco's
    # but we have to crank up solver iterations pretty high.  improving
    # brax's constraint solver will have a significant impact on perf
    sys = sys.replace(solver_iterations=500)
    err, mj_err = 0, 0
    # force PGS so we can reference efc_AR:
    samples = test_utils.sample_mujoco_states(xml_file, force_pgs=True)
    for mj_prev, mj_next in samples:
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)
      state = jax.jit(pipeline.step)(sys, state, mj_prev.qfrc_applied)
      # recover con_frc by backing it out from qf_constraint
      con_frc = np.linalg.lstsq(mj_next.efc_J.T, state.qf_constraint, None)[0]
      err += np.sum((mj_next.efc_AR @ con_frc + mj_next.efc_b) ** 2)
      mj_err += np.sum(
          (mj_next.efc_AR @ mj_next.efc_force + mj_next.efc_b) ** 2
      )

    self.assertLessEqual(err, mj_err + 0.01)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for running statistics utilities.

This file was taken from acme and modified to simplify dependencies:

https://github.com/deepmind/acme/blob/master/acme/jax/running_statistics_test.py
"""

import functools
import math

from absl.testing import absltest
from brax.training.acme import running_statistics
from brax.training.acme import specs
from jax.config import config as jax_config
import jax.numpy as jnp
import numpy as np

update_and_validate = functools.partial(
    running_statistics.update, validate_shapes=True)


class RunningStatisticsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    jax_config.update('jax_enable_x64', False)

  def assert_allclose(self,
                      actual: jnp.ndarray,
                      desired: jnp.ndarray,
                      err_msg: str = '') -> None:
    np.testing.assert_allclose(
        actual, desired, atol=1e-5, rtol=1e-5, err_msg=err_msg)

  def test_normalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(200, dtype=jnp.float32).reshape(20, 2, 5)
    x1, x2, x3, x4 = jnp.split(x, 4, axis=0)

    state = update_and_validate(state, x1)
    state = update_and_validate(state, x2)
    state = update_and_validate(state, x3)
    state = update_and_validate(state, x4)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assert_allclose(mean, jnp.zeros_like(mean))
    self.assert_allclose(std, jnp.ones_like(std))

  def test_init_normalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(200, dtype=jnp.float32).reshape(20, 2, 5)
    normalized = running_statistics.normalize(x, state)

    self.assert_allclose(normalized, x)

  def test_one_batch_dim(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(10, dtype=jnp.float32).reshape(2, 5)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized, axis=0)
    std = jnp.std(normalized, axis=0)
    self.assert_allclose(mean, jnp.zeros_like(mean))
    self.assert_allclose(std, jnp.ones_like(std))

  def test_clip(self):
    state = running_statistics.init_state(specs.Array((), jnp.float32))

    x = jnp.arange(5, dtype=jnp.float32)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state, max_abs_value=1.0)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assert_allclose(mean, jnp.zeros_like(mean))
    self.assert_allclose(std, jnp.ones_like(std) * math.sqrt(0.6))

  def test_validation(self):
    state = running_statistics.init_state(specs.Array((1, 2, 3), jnp.float32))

    x = jnp.arange(12, dtype=jnp.float32).reshape(2, 2, 3)
    with self.assertRaises(AssertionError):
      update_and_validate(state, x)

    x = jnp.arange(3, dtype=jnp.float32).reshape(1, 1, 3)
    with self.assertRaises(AssertionError):
      update_and_validate(state, x)

  def test_int_not_normalized(self):
    state = running_statistics.init_state(specs.Array((), jnp.int32))

    x = jnp.arange(5, dtype=jnp.int32)

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state)

    np.testing.assert_array_equal(normalized, x)

  def test_weights(self):
    state = running_statistics.init_state(specs.Array((), jnp.float32))

    x = jnp.arange(5, dtype=jnp.float32)
    x_weights = jnp.ones_like(x)
    y = 2 * x + 5
    y_weights = 2 * x_weights
    z = jnp.concatenate([x, y])
    weights = jnp.concatenate([x_weights, y_weights])

    state = update_and_validate(state, z, weights=weights)

    self.assertEqual(state.mean, (jnp.mean(x) + 2 * jnp.mean(y)) / 3)
    big_z = jnp.concatenate([x, y, y])
    normalized = running_statistics.normalize(big_z, state)
    self.assertAlmostEqual(jnp.mean(normalized), 0., places=6)
    self.assertAlmostEqual(jnp.std(normalized), 1., places=6)

  def test_denormalize(self):
    state = running_statistics.init_state(specs.Array((5,), jnp.float32))

    x = jnp.arange(100, dtype=jnp.float32).reshape(10, 2, 5)
    x1, x2 = jnp.split(x, 2, axis=0)

    state = update_and_validate(state, x1)
    state = update_and_validate(state, x2)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized)
    std = jnp.std(normalized)
    self.assert_allclose(mean, jnp.zeros_like(mean))
    self.assert_allclose(std, jnp.ones_like(std))

    denormalized = running_statistics.denormalize(normalized, state)
    self.assert_allclose(denormalized, x)

  def test_nest(self):
    state = running_statistics.init_state(
        dict(dummy=specs.Array((5,), jnp.float32)))

    x = dict(dummy=jnp.arange(10, dtype=jnp.float32).reshape(2, 5))

    state = update_and_validate(state, x)
    normalized = running_statistics.normalize(x, state)

    mean = jnp.mean(normalized['dummy'], axis=0)
    std = jnp.std(normalized['dummy'], axis=0)

    self.assert_allclose(mean, jnp.zeros_like(mean))
    self.assert_allclose(std, jnp.ones_like(std))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to compute running statistics.

This file was taken from acme and modified to simplify dependencies:

https://github.com/deepmind/acme/blob/master/acme/jax/running_statistics.py
"""

from typing import Any, Optional, Tuple

from brax.training.acme import types
from flax import struct
import jax
import jax.numpy as jnp


def _zeros_like(nest: types.Nest, dtype=None) -> types.Nest:
  return jax.tree_util.tree_map(lambda x: jnp.zeros(x.shape, dtype or x.dtype), nest)


def _ones_like(nest: types.Nest, dtype=None) -> types.Nest:
  return jax.tree_util.tree_map(lambda x: jnp.ones(x.shape, dtype or x.dtype), nest)


@struct.dataclass
class NestedMeanStd:
  """A container for running statistics (mean, std) of possibly nested data."""
  mean: types.Nest
  std: types.Nest


@struct.dataclass
class RunningStatisticsState(NestedMeanStd):
  """Full state of running statistics computation."""
  count: jnp.ndarray
  summed_variance: types.Nest


def init_state(nest: types.Nest) -> RunningStatisticsState:
  """Initializes the running statistics for the given nested structure."""
  dtype = jnp.float64 if jax.config.jax_enable_x64 else jnp.float32

  return RunningStatisticsState(
      count=jnp.zeros((), dtype=dtype),
      mean=_zeros_like(nest, dtype=dtype),
      summed_variance=_zeros_like(nest, dtype=dtype),
      # Initialize with ones to make sure normalization works correctly
      # in the initial state.
      std=_ones_like(nest, dtype=dtype))


def _validate_batch_shapes(batch: types.NestedArray,
                           reference_sample: types.NestedArray,
                           batch_dims: Tuple[int, ...]) -> None:
  """Verifies shapes of the batch leaves against the reference sample.

  Checks that batch dimensions are the same in all leaves in the batch.
  Checks that non-batch dimensions for all leaves in the batch are the same
  as in the reference sample.

  Arguments:
    batch: the nested batch of data to be verified.
    reference_sample: the nested array to check non-batch dimensions.
    batch_dims: a Tuple of indices of batch dimensions in the batch shape.

  Returns:
    None.
  """
  def validate_node_shape(reference_sample: jnp.ndarray,
                          batch: jnp.ndarray) -> None:
    expected_shape = batch_dims + reference_sample.shape
    assert batch.shape == expected_shape, f'{batch.shape} != {expected_shape}'

  jax.tree_util.tree_map(validate_node_shape, reference_sample, batch)


def update(state: RunningStatisticsState,
           batch: types.Nest,
           *,
           weights: Optional[jnp.ndarray] = None,
           std_min_value: float = 1e-6,
           std_max_value: float = 1e6,
           pmap_axis_name: Optional[str] = None,
           validate_shapes: bool = True) -> RunningStatisticsState:
  """Updates the running statistics with the given batch of data.

  Note: data batch and state elements (mean, etc.) must have the same structure.

  Note: by default will use int32 for counts and float32 for accumulated
  variance. This results in an integer overflow after 2^31 data points and
  degrading precision after 2^24 batch updates or even earlier if variance
  updates have large dynamic range.
  To improve precision, consider setting jax_enable_x64 to True, see
  https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#double-64bit-precision

  Arguments:
    state: The running statistics before the update.
    batch: The data to be used to update the running statistics.
    weights: Weights of the batch data. Should match the batch dimensions.
      Passing a weight of 2. should be equivalent to updating on the
      corresponding data point twice.
    std_min_value: Minimum value for the standard deviation.
    std_max_value: Maximum value for the standard deviation.
    pmap_axis_name: Name of the pmapped axis, if any.
    validate_shapes: If true, the shapes of all leaves of the batch will be
      validated. Enabled by default. Doesn't impact performance when jitted.

  Returns:
    Updated running statistics.
  """
  # We require exactly the same structure to avoid issues when flattened
  # batch and state have different order of elements.
  assert jax.tree_util.tree_structure(batch) == jax.tree_util.tree_structure(state.mean)
  batch_shape = jax.tree_util.tree_leaves(batch)[0].shape
  # We assume the batch dimensions always go first.
  batch_dims = batch_shape[:len(batch_shape) -
                           jax.tree_util.tree_leaves(state.mean)[0].ndim]
  batch_axis = range(len(batch_dims))
  if weights is None:
    step_increment = jnp.prod(jnp.array(batch_dims))
  else:
    step_increment = jnp.sum(weights)
  if pmap_axis_name is not None:
    step_increment = jax.lax.psum(step_increment, axis_name=pmap_axis_name)
  count = state.count + step_increment

  # Validation is important. If the shapes don't match exactly, but are
  # compatible, arrays will be silently broadcasted resulting in incorrect
  # statistics.
  if validate_shapes:
    if weights is not None:
      if weights.shape != batch_dims:
        raise ValueError(f'{weights.shape} != {batch_dims}')
    _validate_batch_shapes(batch, state.mean, batch_dims)

  def _compute_node_statistics(
      mean: jnp.ndarray, summed_variance: jnp.ndarray,
      batch: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    assert isinstance(mean, jnp.ndarray), type(mean)
    assert isinstance(summed_variance, jnp.ndarray), type(summed_variance)
    # The mean and the sum of past variances are updated with Welford's
    # algorithm using batches (see https://stackoverflow.com/q/56402955).
    diff_to_old_mean = batch - mean
    if weights is not None:
      expanded_weights = jnp.reshape(
          weights,
          list(weights.shape) + [1] * (batch.ndim - weights.ndim))
      diff_to_old_mean = diff_to_old_mean * expanded_weights
    mean_update = jnp.sum(diff_to_old_mean, axis=batch_axis) / count
    if pmap_axis_name is not None:
      mean_update = jax.lax.psum(
          mean_update, axis_name=pmap_axis_name)
    mean = mean + mean_update

    diff_to_new_mean = batch - mean
    variance_update = diff_to_old_mean * diff_to_new_mean
    variance_update = jnp.sum(variance_update, axis=batch_axis)
    if pmap_axis_name is not None:
      variance_update = jax.lax.psum(variance_update, axis_name=pmap_axis_name)
    summed_variance = summed_variance + variance_update
    return mean, summed_variance

  updated_stats = jax.tree_util.tree_map(_compute_node_statistics, state.mean,
                                         state.summed_variance, batch)
  # Extract `mean` and `summed_variance` from `updated_stats` nest.
  mean = jax.tree_util.tree_map(lambda _, x: x[0], state.mean, updated_stats)
  summed_variance = jax.tree_util.tree_map(lambda _, x: x[1], state.mean,
                                           updated_stats)

  def compute_std(summed_variance: jnp.ndarray,
                  std: jnp.ndarray) -> jnp.ndarray:
    assert isinstance(summed_variance, jnp.ndarray)
    # Summed variance can get negative due to rounding errors.
    summed_variance = jnp.maximum(summed_variance, 0)
    std = jnp.sqrt(summed_variance / count)
    std = jnp.clip(std, std_min_value, std_max_value)
    return std

  std = jax.tree_util.tree_map(compute_std, summed_variance, state.std)

  return RunningStatisticsState(
      count=count, mean=mean, summed_variance=summed_variance, std=std)


def normalize(batch: types.NestedArray,
              mean_std: NestedMeanStd,
              max_abs_value: Optional[float] = None) -> types.NestedArray:
  """Normalizes data using running statistics."""

  def normalize_leaf(data: jnp.ndarray, mean: jnp.ndarray,
                     std: jnp.ndarray) -> jnp.ndarray:
    # Only normalize inexact
    if not jnp.issubdtype(data.dtype, jnp.inexact):
      return data
    data = (data - mean) / std
    if max_abs_value is not None:
      # TODO: remove pylint directive
      data = jnp.clip(data, -max_abs_value, +max_abs_value)
    return data

  return jax.tree_util.tree_map(normalize_leaf, batch, mean_std.mean, mean_std.std)


def denormalize(batch: types.NestedArray,
                mean_std: NestedMeanStd) -> types.NestedArray:
  """Denormalizes values in a nested structure using the given mean/std.

  Only values of inexact types are denormalized.
  See https://numpy.org/doc/stable/_images/dtype-hierarchy.png for Numpy type
  hierarchy.

  Args:
    batch: a nested structure containing batch of data.
    mean_std: mean and standard deviation used for denormalization.

  Returns:
    Nested structure with denormalized values.
  """

  def denormalize_leaf(data: jnp.ndarray, mean: jnp.ndarray,
                       std: jnp.ndarray) -> jnp.ndarray:
    # Only denormalize inexact
    if not jnp.issubdtype(data.dtype, jnp.inexact):
      return data
    return data * std + mean

  return jax.tree_util.tree_map(denormalize_leaf, batch, mean_std.mean, mean_std.std)




# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loads mesh from disk."""

import os
from typing import Sequence

from brax.io import file
from brax.physics import config_pb2
from trimesh.exchange.load import load_mesh


def load(name: str, path: str,
         resource_paths: Sequence[str]) -> config_pb2.MeshGeometry:
  """Returns MeshGeometry with faces/vertices loaded from path.

  Args:
    name: Name of the mesh.
    path: Path to mesh file name.
    resource_paths: Sequence of paths that may contain the mesh geometry file.

  Returns:
    mesh_geom: MeshGeometry object.

  Raises:
    AssertionError: if the mesh_geom.path is not found in resource_paths.
  """
  mesh_geom = config_pb2.MeshGeometry()
  mesh_geom.name = name
  mesh_geom.path = path

  # Load the first existing mesh file into mesh_geom.
  for resource_path in resource_paths:
    path = os.path.join(resource_path, path)
    if not file.Exists(path):
      continue
    with file.File(path, 'rb') as f:
      mesh = load_mesh(f, file_type=str(mesh_geom.path))
      for v in mesh.vertices:
        mesh_geom.vertices.add(x=v[0], y=v[1], z=v[2])
      mesh_geom.faces.extend(mesh.faces.flatten())
      for v in mesh.vertex_normals:
        mesh_geom.vertex_normals.add(x=v[0], y=v[1], z=v[2])
      for v in mesh.face_normals:
        mesh_geom.face_normals.add(x=v[0], y=v[1], z=v[2])
    return mesh_geom

  raise AssertionError(f'{mesh_geom.path} was not found.')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: brax/physics/config.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x62rax/physics/config.proto\x12\x04\x62rax\"*\n\x07Vector3\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"W\n\x06\x46rozen\x12\x1f\n\x08position\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12\x0b\n\x03\x61ll\x18\x03 \x01(\x08\"\x83\x01\n\x04\x42ody\x12\x0c\n\x04name\x18\x01 \x01(\t\x12!\n\tcolliders\x18\x02 \x03(\x0b\x32\x0e.brax.Collider\x12\x1e\n\x07inertia\x18\x03 \x01(\x0b\x32\r.brax.Vector3\x12\x0c\n\x04mass\x18\x04 \x01(\x02\x12\x1c\n\x06\x66rozen\x18\x05 \x01(\x0b\x32\x0c.brax.Frozen\"\x95\x06\n\x08\x43ollider\x12)\n\x08material\x18\x08 \x01(\x0b\x32\x17.brax.Collider.Material\x12\x1f\n\x08position\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12!\n\x03\x62ox\x18\x03 \x01(\x0b\x32\x12.brax.Collider.BoxH\x00\x12%\n\x05plane\x18\x04 \x01(\x0b\x32\x14.brax.Collider.PlaneH\x00\x12\'\n\x06sphere\x18\x05 \x01(\x0b\x32\x15.brax.Collider.SphereH\x00\x12)\n\x07\x63\x61psule\x18\x06 \x01(\x0b\x32\x16.brax.Collider.CapsuleH\x00\x12-\n\theightMap\x18\x07 \x01(\x0b\x32\x18.brax.Collider.HeightMapH\x00\x12#\n\x04mesh\x18\t \x01(\x0b\x32\x13.brax.Collider.MeshH\x00\x12\x34\n\rclipped_plane\x18\r \x01(\x0b\x32\x1b.brax.Collider.ClippedPlaneH\x00\x12\r\n\x05\x63olor\x18\n \x01(\t\x12\x0e\n\x06hidden\x18\x0b \x01(\x08\x12\x12\n\nno_contact\x18\x0c \x01(\x08\x1a\x30\n\x08Material\x12\x12\n\nelasticity\x18\x01 \x01(\x02\x12\x10\n\x08\x66riction\x18\x02 \x01(\x02\x1a&\n\x03\x42ox\x12\x1f\n\x08halfsize\x18\x01 \x01(\x0b\x32\r.brax.Vector3\x1a\x07\n\x05Plane\x1a\x36\n\x0c\x43lippedPlane\x12\x12\n\nhalfsize_x\x18\x01 \x01(\x02\x12\x12\n\nhalfsize_y\x18\x02 \x01(\x02\x1a\x18\n\x06Sphere\x12\x0e\n\x06radius\x18\x01 \x01(\x02\x1a\x36\n\x07\x43\x61psule\x12\x0e\n\x06radius\x18\x01 \x01(\x02\x12\x0e\n\x06length\x18\x02 \x01(\x02\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x05\x1a\'\n\tHeightMap\x12\x0c\n\x04size\x18\x01 \x01(\x02\x12\x0c\n\x04\x64\x61ta\x18\x02 \x03(\x02\x1a#\n\x04Mesh\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05scale\x18\x02 \x01(\x02\x42\x06\n\x04type\"\xa2\x03\n\x05Joint\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tstiffness\x18\x02 \x01(\x02\x12\x0e\n\x06parent\x18\x03 \x01(\t\x12\r\n\x05\x63hild\x18\x04 \x01(\t\x12$\n\rparent_offset\x18\x05 \x01(\x0b\x32\r.brax.Vector3\x12#\n\x0c\x63hild_offset\x18\x06 \x01(\x0b\x32\r.brax.Vector3\x12\x1f\n\x08rotation\x18\x07 \x01(\x0b\x32\r.brax.Vector3\x12\x17\n\x0f\x61ngular_damping\x18\x08 \x01(\x02\x12&\n\x0b\x61ngle_limit\x18\t \x03(\x0b\x32\x11.brax.Joint.Range\x12\x1b\n\x0elimit_strength\x18\n \x01(\x02H\x00\x88\x01\x01\x12\x1b\n\x0espring_damping\x18\x0b \x01(\x02H\x01\x88\x01\x01\x12)\n\x12reference_rotation\x18\x0c \x01(\x0b\x32\r.brax.Vector3\x1a!\n\x05Range\x12\x0b\n\x03min\x18\x01 \x01(\x02\x12\x0b\n\x03max\x18\x02 \x01(\x02\x42\x11\n\x0f_limit_strengthB\x11\n\x0f_spring_damping\"\xa4\x01\n\x08\x41\x63tuator\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05joint\x18\x02 \x01(\t\x12\x10\n\x08strength\x18\x03 \x01(\x02\x12\'\n\x06torque\x18\x04 \x01(\x0b\x32\x15.brax.Actuator.TorqueH\x00\x12%\n\x05\x61ngle\x18\x05 \x01(\x0b\x32\x14.brax.Actuator.AngleH\x00\x1a\x08\n\x06Torque\x1a\x07\n\x05\x41ngleB\x06\n\x04type\"\xa6\x01\n\x05\x46orce\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\t\x12\x10\n\x08strength\x18\x03 \x01(\x02\x12(\n\x08thruster\x18\x04 \x01(\x0b\x32\x14.brax.Force.ThrusterH\x00\x12&\n\x07twister\x18\x05 \x01(\x0b\x32\x13.brax.Force.TwisterH\x00\x1a\n\n\x08Thruster\x1a\t\n\x07TwisterB\x06\n\x04type\"\xa0\x02\n\x0c\x44\x65\x66\x61ultState\x12-\n\x06\x61ngles\x18\x01 \x03(\x0b\x32\x1d.brax.DefaultState.JointAngle\x12\"\n\x03qps\x18\x02 \x03(\x0b\x32\x15.brax.DefaultState.QP\x1a\x38\n\nJointAngle\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1c\n\x05\x61ngle\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x1a\x82\x01\n\x02QP\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x03pos\x18\x02 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03rot\x18\x03 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03vel\x18\x04 \x01(\x0b\x32\r.brax.Vector3\x12\x1a\n\x03\x61ng\x18\x05 \x01(\x0b\x32\r.brax.Vector3\"\xa6\x01\n\x0cMeshGeometry\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x1f\n\x08vertices\x18\x03 \x03(\x0b\x32\r.brax.Vector3\x12\r\n\x05\x66\x61\x63\x65s\x18\x04 \x03(\x05\x12%\n\x0evertex_normals\x18\x05 \x03(\x0b\x32\r.brax.Vector3\x12#\n\x0c\x66\x61\x63\x65_normals\x18\x06 \x03(\x0b\x32\r.brax.Vector3\"\xcf\x05\n\x06\x43onfig\x12\x1a\n\x06\x62odies\x18\x01 \x03(\x0b\x32\n.brax.Body\x12\x1b\n\x06joints\x18\x02 \x03(\x0b\x32\x0b.brax.Joint\x12!\n\tactuators\x18\x03 \x03(\x0b\x32\x0e.brax.Actuator\x12\x1b\n\x06\x66orces\x18\x10 \x03(\x0b\x32\x0b.brax.Force\x12\x12\n\nelasticity\x18\x04 \x01(\x02\x12\x10\n\x08\x66riction\x18\x05 \x01(\x02\x12\x1e\n\x07gravity\x18\x06 \x01(\x0b\x32\r.brax.Vector3\x12\x18\n\x10velocity_damping\x18\x07 \x01(\x02\x12\x17\n\x0f\x61ngular_damping\x18\x08 \x01(\x02\x12\x15\n\rbaumgarte_erp\x18\t \x01(\x02\x12.\n\x0f\x63ollide_include\x18\n \x03(\x0b\x32\x15.brax.Config.NamePair\x12\n\n\x02\x64t\x18\x0b \x01(\x02\x12\x10\n\x08substeps\x18\x0c \x01(\x05\x12\x1c\n\x06\x66rozen\x18\r \x01(\x0b\x32\x0c.brax.Frozen\x12$\n\x08\x64\x65\x66\x61ults\x18\x0e \x03(\x0b\x32\x12.brax.DefaultState\x12\x17\n\x0f\x63ollider_cutoff\x18\x0f \x01(\x05\x12+\n\x0fmesh_geometries\x18\x11 \x03(\x0b\x32\x12.brax.MeshGeometry\x12\x15\n\rdynamics_mode\x18\x12 \x01(\t\x12\x1d\n\x10solver_scale_pos\x18\x13 \x01(\x02H\x00\x88\x01\x01\x12\x1d\n\x10solver_scale_ang\x18\x14 \x01(\x02H\x01\x88\x01\x01\x12!\n\x14solver_scale_collide\x18\x15 \x01(\x02H\x02\x88\x01\x01\x1a)\n\x08NamePair\x12\r\n\x05\x66irst\x18\x01 \x01(\t\x12\x0e\n\x06second\x18\x02 \x01(\tB\x13\n\x11_solver_scale_posB\x13\n\x11_solver_scale_angB\x17\n\x15_solver_scale_collideb\x06proto3')



_VECTOR3 = DESCRIPTOR.message_types_by_name['Vector3']
_FROZEN = DESCRIPTOR.message_types_by_name['Frozen']
_BODY = DESCRIPTOR.message_types_by_name['Body']
_COLLIDER = DESCRIPTOR.message_types_by_name['Collider']
_COLLIDER_MATERIAL = _COLLIDER.nested_types_by_name['Material']
_COLLIDER_BOX = _COLLIDER.nested_types_by_name['Box']
_COLLIDER_PLANE = _COLLIDER.nested_types_by_name['Plane']
_COLLIDER_CLIPPEDPLANE = _COLLIDER.nested_types_by_name['ClippedPlane']
_COLLIDER_SPHERE = _COLLIDER.nested_types_by_name['Sphere']
_COLLIDER_CAPSULE = _COLLIDER.nested_types_by_name['Capsule']
_COLLIDER_HEIGHTMAP = _COLLIDER.nested_types_by_name['HeightMap']
_COLLIDER_MESH = _COLLIDER.nested_types_by_name['Mesh']
_JOINT = DESCRIPTOR.message_types_by_name['Joint']
_JOINT_RANGE = _JOINT.nested_types_by_name['Range']
_ACTUATOR = DESCRIPTOR.message_types_by_name['Actuator']
_ACTUATOR_TORQUE = _ACTUATOR.nested_types_by_name['Torque']
_ACTUATOR_ANGLE = _ACTUATOR.nested_types_by_name['Angle']
_FORCE = DESCRIPTOR.message_types_by_name['Force']
_FORCE_THRUSTER = _FORCE.nested_types_by_name['Thruster']
_FORCE_TWISTER = _FORCE.nested_types_by_name['Twister']
_DEFAULTSTATE = DESCRIPTOR.message_types_by_name['DefaultState']
_DEFAULTSTATE_JOINTANGLE = _DEFAULTSTATE.nested_types_by_name['JointAngle']
_DEFAULTSTATE_QP = _DEFAULTSTATE.nested_types_by_name['QP']
_MESHGEOMETRY = DESCRIPTOR.message_types_by_name['MeshGeometry']
_CONFIG = DESCRIPTOR.message_types_by_name['Config']
_CONFIG_NAMEPAIR = _CONFIG.nested_types_by_name['NamePair']
Vector3 = _reflection.GeneratedProtocolMessageType('Vector3', (_message.Message,), {
  'DESCRIPTOR' : _VECTOR3,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Vector3)
  })
_sym_db.RegisterMessage(Vector3)

Frozen = _reflection.GeneratedProtocolMessageType('Frozen', (_message.Message,), {
  'DESCRIPTOR' : _FROZEN,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Frozen)
  })
_sym_db.RegisterMessage(Frozen)

Body = _reflection.GeneratedProtocolMessageType('Body', (_message.Message,), {
  'DESCRIPTOR' : _BODY,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Body)
  })
_sym_db.RegisterMessage(Body)

Collider = _reflection.GeneratedProtocolMessageType('Collider', (_message.Message,), {

  'Material' : _reflection.GeneratedProtocolMessageType('Material', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_MATERIAL,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Material)
    })
  ,

  'Box' : _reflection.GeneratedProtocolMessageType('Box', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_BOX,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Box)
    })
  ,

  'Plane' : _reflection.GeneratedProtocolMessageType('Plane', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_PLANE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Plane)
    })
  ,

  'ClippedPlane' : _reflection.GeneratedProtocolMessageType('ClippedPlane', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_CLIPPEDPLANE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.ClippedPlane)
    })
  ,

  'Sphere' : _reflection.GeneratedProtocolMessageType('Sphere', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_SPHERE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Sphere)
    })
  ,

  'Capsule' : _reflection.GeneratedProtocolMessageType('Capsule', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_CAPSULE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Capsule)
    })
  ,

  'HeightMap' : _reflection.GeneratedProtocolMessageType('HeightMap', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_HEIGHTMAP,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.HeightMap)
    })
  ,

  'Mesh' : _reflection.GeneratedProtocolMessageType('Mesh', (_message.Message,), {
    'DESCRIPTOR' : _COLLIDER_MESH,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Collider.Mesh)
    })
  ,
  'DESCRIPTOR' : _COLLIDER,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Collider)
  })
_sym_db.RegisterMessage(Collider)
_sym_db.RegisterMessage(Collider.Material)
_sym_db.RegisterMessage(Collider.Box)
_sym_db.RegisterMessage(Collider.Plane)
_sym_db.RegisterMessage(Collider.ClippedPlane)
_sym_db.RegisterMessage(Collider.Sphere)
_sym_db.RegisterMessage(Collider.Capsule)
_sym_db.RegisterMessage(Collider.HeightMap)
_sym_db.RegisterMessage(Collider.Mesh)

Joint = _reflection.GeneratedProtocolMessageType('Joint', (_message.Message,), {

  'Range' : _reflection.GeneratedProtocolMessageType('Range', (_message.Message,), {
    'DESCRIPTOR' : _JOINT_RANGE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Joint.Range)
    })
  ,
  'DESCRIPTOR' : _JOINT,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Joint)
  })
_sym_db.RegisterMessage(Joint)
_sym_db.RegisterMessage(Joint.Range)

Actuator = _reflection.GeneratedProtocolMessageType('Actuator', (_message.Message,), {

  'Torque' : _reflection.GeneratedProtocolMessageType('Torque', (_message.Message,), {
    'DESCRIPTOR' : _ACTUATOR_TORQUE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Actuator.Torque)
    })
  ,

  'Angle' : _reflection.GeneratedProtocolMessageType('Angle', (_message.Message,), {
    'DESCRIPTOR' : _ACTUATOR_ANGLE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Actuator.Angle)
    })
  ,
  'DESCRIPTOR' : _ACTUATOR,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Actuator)
  })
_sym_db.RegisterMessage(Actuator)
_sym_db.RegisterMessage(Actuator.Torque)
_sym_db.RegisterMessage(Actuator.Angle)

Force = _reflection.GeneratedProtocolMessageType('Force', (_message.Message,), {

  'Thruster' : _reflection.GeneratedProtocolMessageType('Thruster', (_message.Message,), {
    'DESCRIPTOR' : _FORCE_THRUSTER,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Force.Thruster)
    })
  ,

  'Twister' : _reflection.GeneratedProtocolMessageType('Twister', (_message.Message,), {
    'DESCRIPTOR' : _FORCE_TWISTER,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Force.Twister)
    })
  ,
  'DESCRIPTOR' : _FORCE,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Force)
  })
_sym_db.RegisterMessage(Force)
_sym_db.RegisterMessage(Force.Thruster)
_sym_db.RegisterMessage(Force.Twister)

DefaultState = _reflection.GeneratedProtocolMessageType('DefaultState', (_message.Message,), {

  'JointAngle' : _reflection.GeneratedProtocolMessageType('JointAngle', (_message.Message,), {
    'DESCRIPTOR' : _DEFAULTSTATE_JOINTANGLE,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.DefaultState.JointAngle)
    })
  ,

  'QP' : _reflection.GeneratedProtocolMessageType('QP', (_message.Message,), {
    'DESCRIPTOR' : _DEFAULTSTATE_QP,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.DefaultState.QP)
    })
  ,
  'DESCRIPTOR' : _DEFAULTSTATE,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.DefaultState)
  })
_sym_db.RegisterMessage(DefaultState)
_sym_db.RegisterMessage(DefaultState.JointAngle)
_sym_db.RegisterMessage(DefaultState.QP)

MeshGeometry = _reflection.GeneratedProtocolMessageType('MeshGeometry', (_message.Message,), {
  'DESCRIPTOR' : _MESHGEOMETRY,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.MeshGeometry)
  })
_sym_db.RegisterMessage(MeshGeometry)

Config = _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {

  'NamePair' : _reflection.GeneratedProtocolMessageType('NamePair', (_message.Message,), {
    'DESCRIPTOR' : _CONFIG_NAMEPAIR,
    '__module__' : 'brax.physics.config_pb2'
    # @@protoc_insertion_point(class_scope:brax.Config.NamePair)
    })
  ,
  'DESCRIPTOR' : _CONFIG,
  '__module__' : 'brax.physics.config_pb2'
  # @@protoc_insertion_point(class_scope:brax.Config)
  })
_sym_db.RegisterMessage(Config)
_sym_db.RegisterMessage(Config.NamePair)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _VECTOR3._serialized_start=35
  _VECTOR3._serialized_end=77
  _FROZEN._serialized_start=79
  _FROZEN._serialized_end=166
  _BODY._serialized_start=169
  _BODY._serialized_end=300
  _COLLIDER._serialized_start=303
  _COLLIDER._serialized_end=1092
  _COLLIDER_MATERIAL._serialized_start=771
  _COLLIDER_MATERIAL._serialized_end=819
  _COLLIDER_BOX._serialized_start=821
  _COLLIDER_BOX._serialized_end=859
  _COLLIDER_PLANE._serialized_start=861
  _COLLIDER_PLANE._serialized_end=868
  _COLLIDER_CLIPPEDPLANE._serialized_start=870
  _COLLIDER_CLIPPEDPLANE._serialized_end=924
  _COLLIDER_SPHERE._serialized_start=926
  _COLLIDER_SPHERE._serialized_end=950
  _COLLIDER_CAPSULE._serialized_start=952
  _COLLIDER_CAPSULE._serialized_end=1006
  _COLLIDER_HEIGHTMAP._serialized_start=1008
  _COLLIDER_HEIGHTMAP._serialized_end=1047
  _COLLIDER_MESH._serialized_start=1049
  _COLLIDER_MESH._serialized_end=1084
  _JOINT._serialized_start=1095
  _JOINT._serialized_end=1513
  _JOINT_RANGE._serialized_start=1442
  _JOINT_RANGE._serialized_end=1475
  _ACTUATOR._serialized_start=1516
  _ACTUATOR._serialized_end=1680
  _ACTUATOR_TORQUE._serialized_start=1655
  _ACTUATOR_TORQUE._serialized_end=1663
  _ACTUATOR_ANGLE._serialized_start=1665
  _ACTUATOR_ANGLE._serialized_end=1672
  _FORCE._serialized_start=1683
  _FORCE._serialized_end=1849
  _FORCE_THRUSTER._serialized_start=1820
  _FORCE_THRUSTER._serialized_end=1830
  _FORCE_TWISTER._serialized_start=1832
  _FORCE_TWISTER._serialized_end=1841
  _DEFAULTSTATE._serialized_start=1852
  _DEFAULTSTATE._serialized_end=2140
  _DEFAULTSTATE_JOINTANGLE._serialized_start=1951
  _DEFAULTSTATE_JOINTANGLE._serialized_end=2007
  _DEFAULTSTATE_QP._serialized_start=2010
  _DEFAULTSTATE_QP._serialized_end=2140
  _MESHGEOMETRY._serialized_start=2143
  _MESHGEOMETRY._serialized_end=2309
  _CONFIG._serialized_start=2312
  _CONFIG._serialized_end=3031
  _CONFIG_NAMEPAIR._serialized_start=2923
  _CONFIG_NAMEPAIR._serialized_end=2964
# @@protoc_insertion_point(module_scope)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train a ComposerEnv."""
import copy
import datetime
from typing import Dict, Any
from brax.experimental.braxlines import experiments
from brax.experimental.braxlines.common import evaluators
from brax.experimental.braxlines.common import logger_utils
from brax.experimental.braxlines.training import ppo
from brax.experimental.composer import composer
from brax.experimental.composer.training import mappo

TASK_KEYS = (('env_name',),)


def train(train_job_params: Dict[str, Any],
          output_dir: str,
          return_dict: Dict[str, float] = None,
          progress_dict: Dict[str, float] = None,
          env_tag: str = None):
  """Train."""
  del env_tag
  return_dict = return_dict or {}
  progress_dict = progress_dict or {}
  logger_utils.save_config(
      f'{output_dir}/config.txt', train_job_params, verbose=True)
  config = copy.deepcopy(train_job_params)
  output_path = output_dir

  # extra parameters
  env_name = config.pop('env_name', 'ant_run')
  env_params = config.pop('env_params', {})
  desc_edits = config.pop('desc_edits', {})
  seed = config.pop('seed', 0)
  eval_seed = config.pop('eval_seed', 0)
  ppo_params = experiments.defaults.get_ppo_params(env_name, default='ant')
  ppo_params.update(config.pop('ppo_params', {}))
  assert not config, f'unused config: {config}'

  env_fn = composer.create_fn(
      env_name=env_name, desc_edits=desc_edits, **env_params)

  # @title Training the custom env
  log_path = output_path
  if log_path:
    log_path = f'{log_path}/training_curves.csv'
  tab = logger_utils.Tabulator(output_path=log_path, append=False)

  # We determined some reasonable hyperparameters offline and share them here.
  times = [datetime.datetime.now()]
  plotpatterns = ['eval/episode_reward', 'eval/episode_score']

  progress, _, _, _ = experiments.get_progress_fn(
      plotpatterns,
      times,
      tab=tab,
      max_ncols=5,
      xlim=[0, ppo_params['num_timesteps']],
      return_dict=return_dict,
      progress_dict=progress_dict)

  ppo_lib = mappo if env_fn().metadata.agent_groups else ppo
  inference_fn, params, _ = ppo_lib.train(
      environment_fn=env_fn,
      progress_fn=progress,
      seed=seed,
      extra_step_kwargs=False,
      **ppo_params)
  time_to_jit = times[1] - times[0]
  time_to_train = times[-1] - times[1]
  print(f'time to jit: {time_to_jit}')
  print(f'time to train: {time_to_train}')
  print(f'Saved logs to {log_path}')
  return_dict.update(dict(time_to_jit=time_to_jit, time_to_train=time_to_train))

  evaluators.visualize_env(
      env_fn=env_fn,
      inference_fn=inference_fn,
      params=params,
      batch_size=0,
      seed=eval_seed,
      output_path=output_path,
      verbose=True)
  return return_dict



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Octopus."""
from brax.experimental.composer.components.common import upright_term_fn

ROOT = 'octopus'

SYSTEM_CONFIG = """
bodies {
  name: "octopus"
  colliders {
    position {
      z: 0.009999999776482582
    }
    sphere {
      radius: 0.009999999776482582
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_body"
  colliders {
    position {
      z: 0.009999999776482582
    }
    sphere {
      radius: 0.009999999776482582
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_B"
  colliders {
    position {
      x: -0.03700000047683716
      z: 0.11810000240802765
    }
    rotation {
      y: -45.00010681152344
    }
    capsule {
      radius: 0.06199999898672104
      length: 0.23399999737739563
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_0_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_0_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_1_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_1_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_1_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_2_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_2_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_2_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_3_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_3_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_3_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_4_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_4_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_4_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_5_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_5_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_5_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_6_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_6_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_6_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_7_B"
  colliders {
    position {
      y: 0.06830000132322311
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.057999998331069946
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_7_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
bodies {
  name: "octopus_0_7_0_0_B"
  colliders {
    position {
      y: 0.03700000047683716
    }
    rotation {
      x: 89.95437622070312
    }
    capsule {
      radius: 0.02199999988079071
      length: 0.11800000071525574
    }
  }
  inertia {
    x: 1.0
    y: 1.0
    z: 1.0
  }
  mass: 1.0
  frozen {
    position {
    }
    rotation {
    }
  }
}
joints {
  name: "$octopus.octopus_body"
  stiffness: 7500.0
  parent: "octopus"
  child: "octopus_body"
  rotation {
    y: -90.0
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_body"
  child: "octopus_0_B"
  parent_offset {
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_0_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_0_B"
  child: "octopus_0_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_0_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_0_0_B"
  child: "octopus_0_0_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_1_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_1_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: 45.0
  }
}
joints {
  name: "octopus_0_1_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_1_B"
  child: "octopus_0_1_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_1_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_1_0_B"
  child: "octopus_0_1_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_2_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_2_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: 90.0
  }
}
joints {
  name: "octopus_0_2_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_2_B"
  child: "octopus_0_2_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_2_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_2_0_B"
  child: "octopus_0_2_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_3_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_3_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: 135.0
  }
}
joints {
  name: "octopus_0_3_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_3_B"
  child: "octopus_0_3_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_3_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_3_0_B"
  child: "octopus_0_3_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_4_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_4_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: 180.0
  }
}
joints {
  name: "octopus_0_4_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_4_B"
  child: "octopus_0_4_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_4_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_4_0_B"
  child: "octopus_0_4_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_5_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_5_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: -135.0
  }
}
joints {
  name: "octopus_0_5_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_5_B"
  child: "octopus_0_5_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_5_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_5_0_B"
  child: "octopus_0_5_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_6_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_6_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: -90.0
  }
}
joints {
  name: "octopus_0_6_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_6_B"
  child: "octopus_0_6_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_6_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_6_0_B"
  child: "octopus_0_6_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_7_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_B"
  child: "octopus_0_7_B"
  parent_offset {
    z: 0.014999999664723873
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
    z: -45.0
  }
}
joints {
  name: "octopus_0_7_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_7_B"
  child: "octopus_0_7_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
joints {
  name: "octopus_0_7_0_0_J_R100"
  stiffness: 7500.0
  parent: "octopus_0_7_0_B"
  child: "octopus_0_7_0_0_B"
  parent_offset {
    y: 0.07400000095367432
  }
  child_offset {
  }
  rotation {
  }
  angular_damping: 10.0
  angle_limit {
    min: -45.0
    max: 45.0
  }
  limit_strength: 400.0
  spring_damping: 50.0
  reference_rotation {
  }
}
actuators {
  name: "$octopus.octopus_body"
  joint: "$octopus.octopus_body"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_J_R100"
  joint: "octopus_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_0_J_R100"
  joint: "octopus_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_0_0_J_R100"
  joint: "octopus_0_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_0_0_0_J_R100"
  joint: "octopus_0_0_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_1_J_R100"
  joint: "octopus_0_1_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_1_0_J_R100"
  joint: "octopus_0_1_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_1_0_0_J_R100"
  joint: "octopus_0_1_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_2_J_R100"
  joint: "octopus_0_2_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_2_0_J_R100"
  joint: "octopus_0_2_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_2_0_0_J_R100"
  joint: "octopus_0_2_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_3_J_R100"
  joint: "octopus_0_3_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_3_0_J_R100"
  joint: "octopus_0_3_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_3_0_0_J_R100"
  joint: "octopus_0_3_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_4_J_R100"
  joint: "octopus_0_4_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_4_0_J_R100"
  joint: "octopus_0_4_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_4_0_0_J_R100"
  joint: "octopus_0_4_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_5_J_R100"
  joint: "octopus_0_5_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_5_0_J_R100"
  joint: "octopus_0_5_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_5_0_0_J_R100"
  joint: "octopus_0_5_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_6_J_R100"
  joint: "octopus_0_6_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_6_0_J_R100"
  joint: "octopus_0_6_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_6_0_0_J_R100"
  joint: "octopus_0_6_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_7_J_R100"
  joint: "octopus_0_7_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_7_0_J_R100"
  joint: "octopus_0_7_0_J_R100"
  strength: 150.0
  torque {
  }
}
actuators {
  name: "octopus_0_7_0_0_J_R100"
  joint: "octopus_0_7_0_0_J_R100"
  strength: 150.0
  torque {
  }
}
"""

COLLIDES = ('octopus', 'octopus_body', 'octopus_0_B')
for i in range(8):
  COLLIDES = (
      f'octopus_0_{i}_B',
      f'octopus_0_{i}_0_B',
      f'octopus_0_{i}_0_0_B',
  )

DEFAULT_OBSERVERS = ('root_z_joints',)


def get_specs():
  return dict(
      message_str=SYSTEM_CONFIG,
      collides=COLLIDES,
      root=ROOT,
      term_fn=upright_term_fn,
      observers=DEFAULT_OBSERVERS)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Training utilities."""
from jax import numpy as jnp


def zero_fn(params, obs: jnp.ndarray, rng: jnp.ndarray, action_size: int):
  """Output zero actions."""
  del params, rng
  return jnp.zeros(obs.shape[:-1] + (action_size,))




# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a halfcheetah to run in the +x direction."""

import brax
from brax import jumpy as jp
from brax.envs import env


class Halfcheetah(env.Env):



  """
  ### Description

  This environment is based on the work by P. Wawrzyński in
  ["A Cat-Like Robot Real-Time Learning to Run"](http://staff.elka.pw.edu.pl/~pwawrzyn/pub-s/0812_LSCLRR.pdf).

  The HalfCheetah is a 2-dimensional robot consisting of 9 links and 8 joints
  connecting them (including two paws).

  The goal is to apply a torque on the joints to make the cheetah run forward
  (right) as fast as possible, with a positive reward allocated based on the
  distance moved forward and a negative reward allocated for moving backward.

  The torso and head of the cheetah are fixed, and the torque can only be
  applied on the other 6 joints over the front and back thighs (connecting to
  the torso), shins (connecting to the thighs) and feet (connecting to the
  shins).

  ### Action Space

  The agents take a 6-element vector for actions. The action space is a
  continuous `(action, action, action, action, action, action)` all in
  `[-1.0, 1.0]`, where `action` represents the numerical torques applied
  between *links*

  | Num | Action                                  | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|-----------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied on the back thigh rotor  | -1          | 1           | bthigh                         | hinge | torque (N m) |
  | 1   | Torque applied on the back shin rotor   | -1          | 1           | bshin                          | hinge | torque (N m) |
  | 2   | Torque applied on the back foot rotor   | -1          | 1           | bfoot                          | hinge | torque (N m) |
  | 3   | Torque applied on the front thigh rotor | -1          | 1           | fthigh                         | hinge | torque (N m) |
  | 4   | Torque applied on the front shin rotor  | -1          | 1           | fshin                          | hinge | torque (N m) |
  | 5   | Torque applied on the front foot rotor  | -1          | 1           | ffoot                          | hinge | torque (N m) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  cheetah, followed by the velocities of those individual parts (their
  derivatives) with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(18,)` where the elements
  correspond to the following:

  | Num | Observation                          | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|--------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | z-coordinate of the center of mass   | -Inf | Inf | rootx                          | slide | position (m)             |
  | 1   | w-orientation of the front tip       | -Inf | Inf | rooty                          | hinge | angle (rad)              |
  | 2   | y-orientation of the front tip       | -Inf | Inf | rooty                          | hinge | angle (rad)              |
  | 3   | angle of the back thigh rotor        | -Inf | Inf | bthigh                         | hinge | angle (rad)              |
  | 4   | angle of the back shin rotor         | -Inf | Inf | bshin                          | hinge | angle (rad)              |
  | 5   | angle of the back foot rotor         | -Inf | Inf | bfoot                          | hinge | angle (rad)              |
  | 6   | velocity of the tip along the y-axis | -Inf | Inf | fthigh                         | hinge | angle (rad)              |
  | 7   | angular velocity of front tip        | -Inf | Inf | fshin                          | hinge | angle (rad)              |
  | 8   | angular velocity of second rotor     | -Inf | Inf | ffoot                          | hinge | angle (rad)              |
  | 9   | x-coordinate of the front tip        | -Inf | Inf | rootx                          | slide | velocity (m/s)           |
  | 10  | y-coordinate of the front tip        | -Inf | Inf | rootz                          | slide | velocity (m/s)           |
  | 11  | angle of the front tip               | -Inf | Inf | rooty                          | hinge | angular velocity (rad/s) |
  | 12  | angle of the second rotor            | -Inf | Inf | bthigh                         | hinge | angular velocity (rad/s) |
  | 13  | angle of the second rotor            | -Inf | Inf | bshin                          | hinge | angular velocity (rad/s) |
  | 14  | velocity of the tip along the x-axis | -Inf | Inf | bfoot                          | hinge | angular velocity (rad/s) |
  | 15  | velocity of the tip along the y-axis | -Inf | Inf | fthigh                         | hinge | angular velocity (rad/s) |
  | 16  | angular velocity of front tip        | -Inf | Inf | fshin                          | hinge | angular velocity (rad/s) |
  | 17  | angular velocity of second rotor     | -Inf | Inf | ffoot                          | hinge | angular velocity (rad/s) |

  ### Rewards

  The reward consists of two parts:

  - *reward_run*: A reward of moving forward which is measured as
    *(x-coordinate before action - x-coordinate after action)/dt*. *dt* is the
    time between actions - the default *dt = 0.05*. This reward would be
    positive if the cheetah runs forward (right) desired.
  - *reward_ctrl*: A negative reward for penalising the cheetah if it takes
    actions that are too large. It is measured as *-coefficient x
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.1

  ### Starting State

  All observations start in state (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,) with a noise added to the initial
  state for stochasticity. As seen before, the first 8 values in the state are
  positional and the last 9 values are velocity. A uniform noise in the range of
  [-0.1, 0.1] is added to the positional values while a standard normal noise
  with a mean of 0 and standard deviation of 0.1 is added to the initial
  velocity values of all zeros.

  ### Episode Termination

  The episode terminates when the episode length is greater than 1000.

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder
  (or by changing the path to a modified XML file in another folder).

  ```
  env = gym.make('HalfCheetah-v2')
  ```

  v3, v4, and v5 take gym.make kwargs such as ctrl_cost_weight,
  reset_noise_scale etc.

  ```
  env = gym.make('HalfCheetah-v5', ctrl_cost_weight=0.1, ....)

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as xml_file, ctrl_cost_weight,
    reset_noise_scale etc. rgb rendering comes from tracking camera (so agent
    does not run away from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks. Added
    reward_threshold to environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self,
               forward_reward_weight=1.0,
               ctrl_cost_weight=0.1,
               reset_noise_scale=0.1,
               legacy_spring=False,
               exclude_current_positions_from_observation=True,
               **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

    self._forward_reward_weight = forward_reward_weight
    self._ctrl_cost_weight = ctrl_cost_weight
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation
    )

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)

    qpos = self.sys.default_angle() + self._noise(rng1)
    qvel = self._noise(rng2)

    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    obs = self._get_obs(qp, self.sys.info(qp))
    reward, done, zero = jp.zeros(3)
    metrics = {
        'x_position': zero,
        'x_velocity': zero,
        'reward_ctrl': zero,
        'reward_run': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)

    velocity = (qp.pos[0] - state.qp.pos[0]) / self.sys.config.dt
    forward_reward = self._forward_reward_weight * velocity[0]
    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))

    obs = self._get_obs(qp, info)
    reward = forward_reward - ctrl_cost
    state.metrics.update(
        x_position=qp.pos[0, 0],
        x_velocity=velocity[0],
        reward_run=forward_reward,
        reward_ctrl=-ctrl_cost)

    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Observe halfcheetah body position and velocities."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)

    # qpos: position and orientation of the torso and the joint angles
    # TODO: convert rot to just y-ang component
    if self._exclude_current_positions_from_observation:
      qpos = [qp.pos[0, 2:], qp.rot[0, (0, 2)], joint_angle]
    else:
      qpos = [qp.pos[0, (0, 2)], qp.rot[0, (0, 2)], joint_angle]

    # qvel: velocity of the torso and the joint angle velocities
    qvel = [qp.vel[0, (0, 2)], qp.ang[0, 1:2], joint_vel]

    return jp.concatenate(qpos + qvel)

  def _noise(self, rng):
    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    return jp.random_uniform(rng, (self.sys.num_joint_dof,), low, hi)


_SYSTEM_CONFIG = """
  bodies {
    name: "torso"
    colliders {
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 1.0920000076293945
      }
    }
    colliders {
      position {
        x: 0.6000000238418579
        z: 0.10000000149011612
      }
      rotation {
        y: 49.847328186035156
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3919999897480011
      }
    }
    inertia {
      x: 0.9447969794273376
      y: 0.9447969794273376
      z: 0.9447969794273376
    }
    mass: 9.457332611083984
  }
  bodies {
    name: "bthigh"
    colliders {
      position {
        x: 0.10000000149011612
        z: -0.12999999523162842
      }
      rotation {
        x: -180.0
        y: 37.723960876464844
        z: -180.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.38199999928474426
      }
    }
    inertia {
      x: 0.029636280611157417
      y: 0.029636280611157417
      z: 0.029636280611157417
    }
    mass: 2.335526943206787
  }
  bodies {
    name: "bshin"
    colliders {
      position {
        x: -0.14000000059604645
        z: -0.07000000029802322
      }
      rotation {
        x: 180.0
        y: -63.68956756591797
        z: 180.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3919999897480011
      }
    }
    inertia {
      x: 0.032029107213020325
      y: 0.032029107213020325
      z: 0.032029107213020325
    }
    mass: 2.402003049850464
  }
  bodies {
    name: "bfoot"
    colliders {
      position {
        x: 0.029999999329447746
        z: -0.09700000286102295
      }
      rotation {
        y: -15.469860076904297
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.2800000011920929
      }
    }
    inertia {
      x: 0.0117056118324399
      y: 0.0117056118324399
      z: 0.0117056118324399
    }
    mass: 1.6574708223342896
  }
  bodies {
    name: "fthigh"
    colliders {
      position {
        x: -0.07000000029802322
        z: -0.11999999731779099
      }
      rotation {
        y: 29.793806076049805
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3580000102519989
      }
    }
    inertia {
      x: 0.024391336366534233
      y: 0.024391336366534233
      z: 0.024391336366534233
    }
    mass: 2.1759843826293945
  }
  bodies {
    name: "fshin"
    colliders {
      position {
        x: 0.06499999761581421
        z: -0.09000000357627869
      }
      rotation {
        y: -34.37746810913086
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.30399999022483826
      }
    }
    inertia {
      x: 0.014954624697566032
      y: 0.014954624697566032
      z: 0.014954624697566032
    }
    mass: 1.8170133829116821
  }
  bodies {
    name: "ffoot"
    colliders {
      position {
        x: 0.04500000178813934
        z: -0.07000000029802322
      }
      rotation {
        y: -34.37746810913086
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.23199999332427979
      }
    }
    inertia {
      x: 0.006711110472679138
      y: 0.006711110472679138
      z: 0.006711110472679138
    }
    mass: 1.3383854627609253
  }
  bodies {
    name: "floor"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      position { x: 1.0 y: 1.0 z: 1.0 }
      rotation { x: 1.0 y: 1.0 z: 1.0 }
    }
  }
  joints {
    name: "bthigh"
    parent: "torso"
    child: "bthigh"
    parent_offset {
      x: -0.5
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -29.793806076049805
      max: 60.16056823730469
    }
    }
  joints {
    name: "bshin"
    parent: "bthigh"
    child: "bshin"
    parent_offset {
      x: 0.1599999964237213
      z: -0.25
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -44.97718811035156
      max: 44.97718811035156
    }
    }
  joints {
    name: "bfoot"
    parent: "bshin"
    child: "bfoot"
    parent_offset {
      x: -0.2800000011920929
      z: -0.14000000059604645
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -22.918312072753906
      max: 44.97718811035156
    }
    }
  joints {
    name: "fthigh"
    parent: "torso"
    child: "fthigh"
    parent_offset {
      x: 0.5
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -57.295780181884766
      max: 40.1070442199707
    }
    }
  joints {
    name: "fshin"
    parent: "fthigh"
    child: "fshin"
    parent_offset {
      x: -0.14000000059604645
      z: -0.23999999463558197
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -68.75493621826172
      max: 49.847328186035156
    }
    }
  joints {
    name: "ffoot"
    parent: "fshin"
    child: "ffoot"
    parent_offset {
      x: 0.12999999523162842
      z: -0.18000000715255737
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angle_limit {
      min: -28.647890090942383
      max: 28.647890090942383
    }
    }
  actuators {
    name: "bthigh"
    joint: "bthigh"
    strength: 120.0
    torque {
    }
  }
  actuators {
    name: "bshin"
    joint: "bshin"
    strength: 90.0
    torque {
    }
  }
  actuators {
    name: "bfoot"
    joint: "bfoot"
    strength: 60
    torque {
    }
  }
  actuators {
    name: "fthigh"
    joint: "fthigh"
    strength: 120.0
    torque {
    }
  }
  actuators {
    name: "fshin"
    joint: "fshin"
    strength: 60
    torque {
    }
  }
  actuators {
    name: "ffoot"
    joint: "ffoot"
    strength: 30.0
    torque {
    }
  }
  friction: 0.77459666924
  gravity {
    z: -9.8100004196167
  }
  angular_damping: -0.009999999776482582
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "bfoot"
  }
  collide_include {
    first: "floor"
    second: "ffoot"
  }
  collide_include {
    first: "floor"
    second: "bthigh"
  }
  collide_include {
    first: "floor"
    second: "fthigh"
  }
  collide_include {
    first: "floor"
    second: "bshin"
  }
  collide_include {
    first: "floor"
    second: "fshin"
  }
  collide_include {
    first: "bfoot"
    second: "ffoot"
  }
  dt: 0.05
  substeps: 16
  frozen {
    position {
      y: 1.0
    }
    rotation {
      x: 1.0
      z: 1.0
    }
  }
  dynamics_mode: "pbd"
"""

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "torso"
    colliders {
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 1.0920000076293945
      }
    }
    colliders {
      position {
        x: 0.6000000238418579
        z: 0.10000000149011612
      }
      rotation {
        y: 49.847328186035156
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3919999897480011
      }
    }
    inertia {
      x: 0.9447969794273376
      y: 0.9447969794273376
      z: 0.9447969794273376
    }
    mass: 9.457332611083984
  }
  bodies {
    name: "bthigh"
    colliders {
      position {
        x: 0.10000000149011612
        z: -0.12999999523162842
      }
      rotation {
        x: -180.0
        y: 37.723960876464844
        z: -180.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.38199999928474426
      }
    }
    inertia {
      x: 0.029636280611157417
      y: 0.029636280611157417
      z: 0.029636280611157417
    }
    mass: 2.335526943206787
  }
  bodies {
    name: "bshin"
    colliders {
      position {
        x: -0.14000000059604645
        z: -0.07000000029802322
      }
      rotation {
        x: 180.0
        y: -63.68956756591797
        z: 180.0
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3919999897480011
      }
    }
    inertia {
      x: 0.032029107213020325
      y: 0.032029107213020325
      z: 0.032029107213020325
    }
    mass: 2.402003049850464
  }
  bodies {
    name: "bfoot"
    colliders {
      position {
        x: 0.029999999329447746
        z: -0.09700000286102295
      }
      rotation {
        y: -15.469860076904297
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.2800000011920929
      }
    }
    inertia {
      x: 0.0117056118324399
      y: 0.0117056118324399
      z: 0.0117056118324399
    }
    mass: 1.6574708223342896
  }
  bodies {
    name: "fthigh"
    colliders {
      position {
        x: -0.07000000029802322
        z: -0.11999999731779099
      }
      rotation {
        y: 29.793806076049805
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.3580000102519989
      }
    }
    inertia {
      x: 0.024391336366534233
      y: 0.024391336366534233
      z: 0.024391336366534233
    }
    mass: 2.1759843826293945
  }
  bodies {
    name: "fshin"
    colliders {
      position {
        x: 0.06499999761581421
        z: -0.09000000357627869
      }
      rotation {
        y: -34.37746810913086
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.30399999022483826
      }
    }
    inertia {
      x: 0.014954624697566032
      y: 0.014954624697566032
      z: 0.014954624697566032
    }
    mass: 1.8170133829116821
  }
  bodies {
    name: "ffoot"
    colliders {
      position {
        x: 0.04500000178813934
        z: -0.07000000029802322
      }
      rotation {
        y: -34.37746810913086
      }
      capsule {
        radius: 0.04600000008940697
        length: 0.23199999332427979
      }
    }
    inertia {
      x: 0.006711110472679138
      y: 0.006711110472679138
      z: 0.006711110472679138
    }
    mass: 1.3383854627609253
  }
  bodies {
    name: "floor"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      position { x: 1.0 y: 1.0 z: 1.0 }
      rotation { x: 1.0 y: 1.0 z: 1.0 }
    }
  }
  joints {
    name: "bthigh"
    stiffness: 25000
    parent: "torso"
    child: "bthigh"
    parent_offset {
      x: -0.5
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 60
    angle_limit {
      min: -29.793806076049805
      max: 60.16056823730469
    }
    limit_strength: 1000
    }
  joints {
    name: "bshin"
    stiffness: 25000
    parent: "bthigh"
    child: "bshin"
    parent_offset {
      x: 0.1599999964237213
      z: -0.25
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 60
    angle_limit {
      min: -44.97718811035156
      max: 44.97718811035156
    }
    limit_strength: 1000
    }
  joints {
    name: "bfoot"
    stiffness: 25000
    parent: "bshin"
    child: "bfoot"
    parent_offset {
      x: -0.2800000011920929
      z: -0.14000000059604645
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 60
    angle_limit {
      min: -22.918312072753906
      max: 44.97718811035156
    }
    limit_strength: 1000
    }
  joints {
    name: "fthigh"
    stiffness: 25000
    parent: "torso"
    child: "fthigh"
    parent_offset {
      x: 0.5
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 60
    angle_limit {
      min: -57.295780181884766
      max: 40.1070442199707
    }
    limit_strength: 1000
    }
  joints {
    name: "fshin"
    stiffness: 25000
    parent: "fthigh"
    child: "fshin"
    parent_offset {
      x: -0.14000000059604645
      z: -0.23999999463558197
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 80.0
    angle_limit {
      min: -68.75493621826172
      max: 49.847328186035156
    }
    limit_strength: 1000
    }
  joints {
    name: "ffoot"
    stiffness: 25000
    parent: "fshin"
    child: "ffoot"
    parent_offset {
      x: 0.12999999523162842
      z: -0.18000000715255737
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    spring_damping: 50.0
    angle_limit {
      min: -28.647890090942383
      max: 28.647890090942383
    }
    limit_strength: 1000
    }
  actuators {
    name: "bthigh"
    joint: "bthigh"
    strength: 120.0
    torque {
    }
  }
  actuators {
    name: "bshin"
    joint: "bshin"
    strength: 90.0
    torque {
    }
  }
  actuators {
    name: "bfoot"
    joint: "bfoot"
    strength: 60
    torque {
    }
  }
  actuators {
    name: "fthigh"
    joint: "fthigh"
    strength: 120.0
    torque {
    }
  }
  actuators {
    name: "fshin"
    joint: "fshin"
    strength: 60
    torque {
    }
  }
  actuators {
    name: "ffoot"
    joint: "ffoot"
    strength: 30.0
    torque {
    }
  }
  friction: 0.77459666924
  gravity {
    z: -9.8100004196167
  }
  angular_damping: -0.009999999776482582
  baumgarte_erp: 0.20000000149011612
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "bfoot"
  }
  collide_include {
    first: "floor"
    second: "ffoot"
  }
  collide_include {
    first: "floor"
    second: "bthigh"
  }
  collide_include {
    first: "floor"
    second: "fthigh"
  }
  collide_include {
    first: "floor"
    second: "bshin"
  }
  collide_include {
    first: "floor"
    second: "fshin"
  }
  collide_include {
    first: "bfoot"
    second: "ffoot"
  }
  dt: 0.05
  substeps: 16
  frozen {
    position {
      y: 1.0
    }
    rotation {
      x: 1.0
      z: 1.0
    }
  }
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""composer Sweep."""
from brax.experimental.braxlines.experiments import defaults

seed = 0

AGENT_MODULE = 'brax.experimental.composer.train'
CONFIG = [
    dict(
        env_name=['pro_ant_run'],
        seed=seed,
        env_params={
            'num_legs': [2, 3, 4, 8, 10],
        },
        ppo_params=defaults.get_ppo_params('ant', 10)),
    dict(
        env_name=['ant_push'],
        seed=seed,
        desc_edits={
            'components.cap1.reward_fns.goal.scale': [0.2, 1, 0.5],
        },
        ppo_params=defaults.get_ppo_params('ant', 10)),
    dict(
        env_name=['ant_chase'],
        seed=seed,
        desc_edits={
            'components.agent2.reward_fns.goal.scale': [0.2, 1, 0.5],
        },
        ppo_params=defaults.get_ppo_params('ant', 10)),
]



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Geometry functions."""

from typing import Tuple
from brax.v2 import math
from jax import numpy as jp


def closest_segment_point(
    a: jp.ndarray, b: jp.ndarray, pt: jp.ndarray
) -> jp.ndarray:
  """Returns the closest point on the a-b line segment to a point pt."""
  ab = b - a
  t = jp.dot(pt - a, ab) / (jp.dot(ab, ab) + 1e-6)
  return a + jp.clip(t, 0.0, 1.0) * ab


def closest_segment_point_and_dist(
    a: jp.ndarray, b: jp.ndarray, pt: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Returns closest point on the line segment and the distance squared."""
  closest = closest_segment_point(a, b, pt)
  dist = (pt - closest).dot(pt - closest)
  return closest, dist


def closest_segment_to_segment_points(
    a0: jp.ndarray, a1: jp.ndarray, b0: jp.ndarray, b1: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Returns closest points on two line segments."""
  # Gets the closest segment points by first finding the closest points
  # between two lines. Points are then clipped to be on the line segments
  # and edge cases with clipping are handled.
  dir_a, len_a = math.normalize(a1 - a0)
  dir_b, len_b = math.normalize(b1 - b0)

  # Segment mid-points.
  half_len_a = len_a * 0.5
  half_len_b = len_b * 0.5
  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Translation between two segment mid-points.
  trans = a_mid - b_mid

  # Parametrize points on each line as follows:
  #  point_on_a = a_mid + t_a * dir_a
  #  point_on_b = b_mid + t_b * dir_b
  # and analytically minimize the distance between the two points.
  dira_dot_dirb = dir_a.dot(dir_b)
  dira_dot_trans = dir_a.dot(trans)
  dirb_dot_trans = dir_b.dot(trans)
  denom = 1 - dira_dot_dirb * dira_dot_dirb

  orig_t_a = (-dira_dot_trans + dira_dot_dirb * dirb_dot_trans) / (denom + 1e-6)
  orig_t_b = dirb_dot_trans + orig_t_a * dira_dot_dirb
  t_a = jp.clip(orig_t_a, -half_len_a, half_len_a)
  t_b = jp.clip(orig_t_b, -half_len_b, half_len_b)

  best_a = a_mid + dir_a * t_a
  best_b = b_mid + dir_b * t_b

  # Resolve edge cases where both closest points are clipped to the segment
  # endpoints by recalculating the closest segment points for the current
  # clipped points, and then picking the pair of points with smallest
  # distance. An example of this edge case is when lines intersect but line
  # segments don't.
  new_a, d1 = closest_segment_point_and_dist(a0, a1, best_b)
  new_b, d2 = closest_segment_point_and_dist(b0, b1, best_a)
  best_a = jp.where(d1 < d2, new_a, best_a)
  best_b = jp.where(d1 < d2, best_b, new_b)

  return best_a, best_b


def closest_segment_point_plane(
    a: jp.ndarray, b: jp.ndarray, p0: jp.ndarray, plane_normal: jp.ndarray
) -> jp.ndarray:
  """Gets the closest point between a line segment and a plane."""
  # If a line segment is parametrized as S(t) = a + t * (b - a), we can
  # plug it into the plane equation dot(n, S(t)) - d = 0, then solve for t to
  # get the line-plane intersection. We then clip t to be in [0, 1] to be on
  # the line segment.
  n = plane_normal
  d = jp.dot(p0, n)  # shortest distance from origin to plane
  denom = jp.dot(n, b - a)
  t = (d - jp.dot(n, a)) / (denom + 1e-6 * (denom == 0.0))
  t = jp.clip(t, 0, 1)
  segment_point = a + t * (b - a)

  return segment_point


def closest_triangle_point(
    p0: jp.ndarray, p1: jp.ndarray, p2: jp.ndarray, pt: jp.ndarray
) -> jp.ndarray:
  """Gets the closest point on a triangle to another point in space."""
  # Parametrize the triangle s.t. a point inside the triangle is
  # Q = p0 + u * e0 + v * e1, when 0 <= u <= 1, 0 <= v <= 1, and
  # 0 <= u + v <= 1. Let e0 = (p1 - p0) and e1 = (p2 - p0).
  # We analytically minimize the distance between the point pt and Q.
  e0 = p1 - p0
  e1 = p2 - p0
  a = e0.dot(e0)
  b = e0.dot(e1)
  c = e1.dot(e1)
  d = pt - p0
  # The determinant is 0 only if the angle between e1 and e0 is 0
  # (i.e. the triangle has overlapping lines).
  det = a * c - b * b
  u = (c * e0.dot(d) - b * e1.dot(d)) / det
  v = (-b * e0.dot(d) + a * e1.dot(d)) / det
  inside = (0 <= u) & (u <= 1) & (0 <= v) & (v <= 1) & (u + v <= 1)
  closest_p = p0 + u * e0 + v * e1
  d0 = (closest_p - pt).dot(closest_p - pt)

  # If the closest point is outside the triangle, it must be on an edge, so we
  # check each triangle edge for a closest point to the point pt.
  closest_p1, d1 = closest_segment_point_and_dist(p0, p1, pt)
  closest_p = jp.where((d0 < d1) & inside, closest_p, closest_p1)
  min_d = jp.where((d0 < d1) & inside, d0, d1)

  closest_p2, d2 = closest_segment_point_and_dist(p1, p2, pt)
  closest_p = jp.where(d2 < min_d, closest_p2, closest_p)
  min_d = jp.minimum(min_d, d2)

  closest_p3, d3 = closest_segment_point_and_dist(p2, p0, pt)
  closest_p = jp.where(d3 < min_d, closest_p3, closest_p)

  return closest_p


def closest_segment_triangle_points(
    a: jp.ndarray,
    b: jp.ndarray,
    p0: jp.ndarray,
    p1: jp.ndarray,
    p2: jp.ndarray,
    triangle_normal: jp.ndarray,
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Gets the closest points between a line segment and triangle."""
  # The closest triangle point is either on the edge or within the triangle.
  # First check triangle edges for the closest point.
  seg_pt1, tri_pt1 = closest_segment_to_segment_points(a, b, p0, p1)
  d1 = (seg_pt1 - tri_pt1).dot(seg_pt1 - tri_pt1)
  seg_pt2, tri_pt2 = closest_segment_to_segment_points(a, b, p1, p2)
  d2 = (seg_pt2 - tri_pt2).dot(seg_pt2 - tri_pt2)
  seg_pt3, tri_pt3 = closest_segment_to_segment_points(a, b, p0, p2)
  d3 = (seg_pt3 - tri_pt3).dot(seg_pt3 - tri_pt3)

  # Next, handle the case where the closest triangle point is inside the
  # triangle. Either the line segment intersects the triangle or a segment
  # endpoint is closest to a point inside the triangle.
  # If the line overlaps the triangle and is parallel to the triangle plane,
  # the chosen triangle point is arbitrary.
  seg_pt4 = closest_segment_point_plane(a, b, p0, triangle_normal)
  tri_pt4 = closest_triangle_point(p0, p1, p2, seg_pt4)
  d4 = (seg_pt4 - tri_pt4).dot(seg_pt4 - tri_pt4)

  # Get the point with minimum distance from the line segment point to the
  # triangle point.
  distance = jp.array([[d1, d2, d3, d4]])
  min_dist = jp.amin(distance)
  mask = (distance == min_dist).T
  seg_pt = jp.array([seg_pt1, seg_pt2, seg_pt3, seg_pt4]) * mask
  tri_pt = jp.array([tri_pt1, tri_pt2, tri_pt3, tri_pt4]) * mask
  seg_pt = jp.sum(seg_pt, axis=0) / jp.sum(mask)
  tri_pt = jp.sum(tri_pt, axis=0) / jp.sum(mask)

  return seg_pt, tri_pt



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for dynamics."""

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import test_utils
from brax.v2.generalized import pipeline
import jax
import numpy as np


class DynamicsTest(parameterized.TestCase):

  @parameterized.parameters(
      'ant.xml', 'triple_pendulum.xml', ('humanoid.xml',), ('halfcheetah.xml',)
  )
  def test_transform_com(self, xml_file):
    """Test dynamics transform com."""
    sys = test_utils.load_fixture(xml_file)
    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)

      np.testing.assert_almost_equal(state.com, mj_next.subtree_com[0], 5)
      mj_cinr_i = np.zeros((state.cinr.i.shape[0], 3, 3))
      mj_cinr_i[:, [0, 1, 2], [0, 1, 2]] = mj_next.cinert[1:, 0:3]  # diagonal
      mj_cinr_i[:, [0, 0, 1], [1, 2, 2]] = mj_next.cinert[1:, 3:6]  # upper tri
      mj_cinr_i[:, [1, 2, 2], [0, 0, 1]] = mj_next.cinert[1:, 3:6]  # lower tri
      mj_cinr_pos = mj_next.cinert[1:, 6:9]

      np.testing.assert_almost_equal(state.cinr.i, mj_cinr_i, 5)
      np.testing.assert_almost_equal(state.cinr.transform.pos, mj_cinr_pos, 5)
      np.testing.assert_almost_equal(state.cinr.mass, mj_next.cinert[1:, 9], 6)
      np.testing.assert_almost_equal(state.cd.matrix(), mj_next.cvel[1:], 5)
      np.testing.assert_almost_equal(state.cdof.matrix(), mj_next.cdof, 6)
      np.testing.assert_almost_equal(state.cdofd.matrix(), mj_next.cdof_dot, 5)

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('halfcheetah.xml',),
  )
  def test_forward(self, xml_file):
    """Test dynamics forward."""
    sys = test_utils.load_fixture(xml_file)
    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      state = jax.jit(pipeline.init)(sys, mj_prev.qpos, mj_prev.qvel)
      state = jax.jit(pipeline.step)(sys, state, mj_prev.qfrc_applied)

      np.testing.assert_allclose(
          state.qf_smooth, mj_next.qfrc_smooth, rtol=1e-5, atol=1e-4
      )


if __name__ == '__main__':
  absltest.main()
